
/* $Id: sdl_v.c 8174 2007-01-17 00:20:23Z Darkvater $ */

#include "../stdafx.h"

#ifdef WITH_SDL

#include "../openttd.h"
#include "../debug.h"
#include "../functions.h"
#include "../gfx.h"
#include "../macros.h"
#include "../sdl.h"
#include "../window.h"
#include "../network.h"
#include "../variables.h"
#include "../screenshot.h"
#include "../table/sprites.h"
#include "../table/strings.h"
#include "sdl_v.h"
#include <SDL.h>
//#include <sipapi.h>

//DKS had to add:
#include <stdarg.h>
#include <SDL_gp2x.h>

static bool _all_modes;

#define MAX_DIRTY_RECTS 100
static SDL_Rect _dirty_rects[MAX_DIRTY_RECTS+1];
static int _num_dirty_rects = MAX_DIRTY_RECTS;

static SDL_Surface *_hw_screen;
static SDL_Surface *_sw_screen;
static unsigned int _width, _height;
bool _stretched = false, _rotated = false;
bool _keymove = false;
bool _textinput_enabled = false;
bool _force_screenupdate = false;
static float zoom_x = 1, zoom_y = 1;

//DKS - this is now variable
//static const unsigned int scrollrate = 5;
static unsigned int scrollrate = 3;

//DKS - new GP2X stuff:
static bool _want_left_handed = false; // was a command-line parameter forcing left-handed used?
static bool _want_F100 = false;		// was a command-line parameter forcing use of F100-style controls used?
static bool _f200_controls = true; //are we using the new F200-style control scheme?
static bool _right_handed = true;  // are we using the right-handed new scheme or left-handed new scheme?
static bool _ext_mouse = false;	   // are we using an external USB mouse?
static int	_fast_cursorrate = 12;	// fast cursor movement rate defaults to 12, can be adjusted by command line parameter
static int	_slow_cursorrate = 3;	// slow cursor movement rate defaults to 3, can be adjusted by command line parameter
bool windowmove_pressed = false;	// have we pressed the control-rightclick hotkey to move window under cursor? (USED IN ../window.c
static bool _controls_swapped = false;	// did GP2X user request swapping of CONTROL/SHIFT on R/L to instead be on VOL DOWN/UP?

SDL_Joystick* joystick;
#define GP2X_UPRIGHT	7
#define GP2X_RIGHT		6
#define GP2X_DOWNRIGHT	5
#define GP2X_DOWN		4
#define GP2X_DOWNLEFT	3
#define GP2X_LEFT		2
#define GP2X_UPLEFT		1
#define GP2X_UP			0
#define GP2X_VOLDOWN	17
#define GP2X_VOLUP		16
#define GP2X_START		8
#define GP2X_SELECT		9
#define GP2X_L			10
#define GP2X_R			11
#define GP2X_A			12
#define GP2X_Y			15
#define GP2X_B			13
#define GP2X_X			14
#define GP2X_PUSH		18

//DKS - new function called from openttd.c if command line parameter is passed requesting F100-style controls
void sdl_setF100(void) 
{
	_want_F100 = true;
}
//DKS - new function called from openttd.c if command line parameter is passed requesting left-handed F200 controls
void sdl_setlefthanded(void)
{
	_want_left_handed = true;
}

// DKS - new function added because when in F200 mode, some users like to mouseclick with L/R instead of select/start of volup/down.
//	... This function gets called if -X command line parameter is passed in launch script.  Causes L/R to be mouseclick and 
//          other changes mentioned above.
void sdl_setswapped(void)
{
	_controls_swapped = true;
}	

//DKS - new functions added to allow user configuration of slow and fast cursor movement rates (for when using DPAD/stick) by
//	passing command line parameter for each
void sdl_setslowcursorrate(int rate)
{
	if ((rate > 0) && (rate < 30)) {
		_slow_cursorrate = rate;
	}
}
void sdl_setfastcursorrate(int rate)
{
	if ((rate > 0) && (rate < 30)) {
		_fast_cursorrate = rate;
	}
}

//DKS - new function called from openttd.c if command line parameter is passed requesting swapping of VOLUP/VOLDOWN mouse clicks to be L/R (or from SELECT/START mouse clicks to L/R if on left-handed F200 setting).  Some users prefer this.  This only applies to F200 style controls, L/R are already used for mouse-clicks on F100
static int GetAsyncKeyState(int key)
{
	switch(key)
	{
	case GP2X_UP:
		if( SDL_JoystickGetButton(joystick, GP2X_UP) ) return -1;
		if( SDL_JoystickGetButton(joystick, GP2X_UPRIGHT) ) return -1;
		if( SDL_JoystickGetButton(joystick, GP2X_UPLEFT) ) return -1;
		return 0;
	case GP2X_DOWN:
		if( SDL_JoystickGetButton(joystick, GP2X_DOWN) ) return -1;
		if( SDL_JoystickGetButton(joystick, GP2X_DOWNRIGHT) ) return -1;
		if( SDL_JoystickGetButton(joystick, GP2X_DOWNLEFT) ) return -1;
		return 0;
	case GP2X_LEFT:
		if( SDL_JoystickGetButton(joystick, GP2X_LEFT) ) return -1;
		if( SDL_JoystickGetButton(joystick, GP2X_UPLEFT) ) return -1;
		if( SDL_JoystickGetButton(joystick, GP2X_DOWNLEFT) ) return -1;
		return 0;
	case GP2X_RIGHT:
		if( SDL_JoystickGetButton(joystick, GP2X_RIGHT) ) return -1;
		if( SDL_JoystickGetButton(joystick, GP2X_UPRIGHT) ) return -1;
		if( SDL_JoystickGetButton(joystick, GP2X_DOWNRIGHT) ) return -1;
		return 0;
	default:
		if( SDL_JoystickGetButton(joystick, key) ) return -1;
		else return 0;
	}	
	return 0;
}

/* The font is generated from Xorg 6x10-L1.bdf and adapted from Minimal Lib by rlyeh */
typedef struct sdl_font        { int x,y,w,wmask,h,fg,bg,solid; unsigned char *data; } sdl_font;
static unsigned char sdl_fontf[256][10] = {
{ 0x00>>2, 0xA8>>2, 0x00>>2, 0x88>>2, 0x00>>2, 0x88>>2, 0x00>>2, 0xA8>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x20>>2, 0x70>>2, 0xF8>>2, 0x70>>2, 0x20>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0xA8>>2, 0x54>>2, 0xA8>>2, 0x54>>2, 0xA8>>2, 0x54>>2, 0xA8>>2, 0x54>>2, 0xA8>>2, 0x54>>2, },
{ 0x00>>2, 0x90>>2, 0x90>>2, 0xF0>>2, 0x90>>2, 0x90>>2, 0x78>>2, 0x10>>2, 0x10>>2, 0x10>>2, },
{ 0x00>>2, 0xE0>>2, 0x80>>2, 0xC0>>2, 0x80>>2, 0xB8>>2, 0x20>>2, 0x30>>2, 0x20>>2, 0x20>>2, },
{ 0x00>>2, 0x70>>2, 0x80>>2, 0x80>>2, 0x70>>2, 0x70>>2, 0x48>>2, 0x70>>2, 0x48>>2, 0x48>>2, },
{ 0x00>>2, 0x80>>2, 0x80>>2, 0x80>>2, 0xF0>>2, 0x78>>2, 0x40>>2, 0x70>>2, 0x40>>2, 0x40>>2, },
{ 0x00>>2, 0x20>>2, 0x50>>2, 0x20>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x20>>2, 0x20>>2, 0xF8>>2, 0x20>>2, 0x20>>2, 0xF8>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x90>>2, 0xD0>>2, 0xD0>>2, 0xB0>>2, 0x90>>2, 0x40>>2, 0x40>>2, 0x40>>2, 0x78>>2, },
{ 0x00>>2, 0x90>>2, 0x90>>2, 0x60>>2, 0x40>>2, 0x78>>2, 0x10>>2, 0x10>>2, 0x10>>2, 0x10>>2, },
{ 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0xE0>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0xE0>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x3C>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, },
{ 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x3C>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0xFC>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, },
{ 0xFC>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0xFC>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0xFF>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0xFC>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0xFC>>2, },
{ 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x3C>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, },
{ 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0xE0>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, },
{ 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0xFC>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0xFC>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, },
{ 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, },
{ 0x00>>2, 0x18>>2, 0x60>>2, 0x80>>2, 0x60>>2, 0x18>>2, 0x00>>2, 0xF8>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0xC0>>2, 0x30>>2, 0x08>>2, 0x30>>2, 0xC0>>2, 0x00>>2, 0xF8>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0xF8>>2, 0x50>>2, 0x50>>2, 0x50>>2, 0x50>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x08>>2, 0x10>>2, 0xF8>>2, 0x20>>2, 0xF8>>2, 0x40>>2, 0x80>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x30>>2, 0x48>>2, 0x40>>2, 0xE0>>2, 0x40>>2, 0x48>>2, 0xB0>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x20>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x00>>2, 0x20>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x50>>2, 0x50>>2, 0x50>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x50>>2, 0x50>>2, 0xF8>>2, 0x50>>2, 0xF8>>2, 0x50>>2, 0x50>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x20>>2, 0x70>>2, 0xA0>>2, 0x70>>2, 0x28>>2, 0x70>>2, 0x20>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x48>>2, 0xA8>>2, 0x50>>2, 0x20>>2, 0x50>>2, 0xA8>>2, 0x90>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x40>>2, 0xA0>>2, 0xA0>>2, 0x40>>2, 0xA8>>2, 0x90>>2, 0x68>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x10>>2, 0x20>>2, 0x40>>2, 0x40>>2, 0x40>>2, 0x20>>2, 0x10>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x40>>2, 0x20>>2, 0x10>>2, 0x10>>2, 0x10>>2, 0x20>>2, 0x40>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x88>>2, 0x50>>2, 0xF8>>2, 0x50>>2, 0x88>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x20>>2, 0x20>>2, 0xF8>>2, 0x20>>2, 0x20>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x30>>2, 0x20>>2, 0x40>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0xF8>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x20>>2, 0x70>>2, 0x20>>2, 0x00>>2, },
{ 0x00>>2, 0x08>>2, 0x08>>2, 0x10>>2, 0x20>>2, 0x40>>2, 0x80>>2, 0x80>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x20>>2, 0x50>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x50>>2, 0x20>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x20>>2, 0x60>>2, 0xA0>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0xF8>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x70>>2, 0x88>>2, 0x08>>2, 0x30>>2, 0x40>>2, 0x80>>2, 0xF8>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0xF8>>2, 0x08>>2, 0x10>>2, 0x30>>2, 0x08>>2, 0x88>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x10>>2, 0x30>>2, 0x50>>2, 0x90>>2, 0xF8>>2, 0x10>>2, 0x10>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0xF8>>2, 0x80>>2, 0xB0>>2, 0xC8>>2, 0x08>>2, 0x88>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x30>>2, 0x40>>2, 0x80>>2, 0xB0>>2, 0xC8>>2, 0x88>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0xF8>>2, 0x08>>2, 0x10>>2, 0x10>>2, 0x20>>2, 0x40>>2, 0x40>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x70>>2, 0x88>>2, 0x88>>2, 0x70>>2, 0x88>>2, 0x88>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x70>>2, 0x88>>2, 0x98>>2, 0x68>>2, 0x08>>2, 0x10>>2, 0x60>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x20>>2, 0x70>>2, 0x20>>2, 0x00>>2, 0x20>>2, 0x70>>2, 0x20>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x20>>2, 0x70>>2, 0x20>>2, 0x00>>2, 0x30>>2, 0x20>>2, 0x40>>2, 0x00>>2, },
{ 0x00>>2, 0x08>>2, 0x10>>2, 0x20>>2, 0x40>>2, 0x20>>2, 0x10>>2, 0x08>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0xF8>>2, 0x00>>2, 0xF8>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x40>>2, 0x20>>2, 0x10>>2, 0x08>>2, 0x10>>2, 0x20>>2, 0x40>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x70>>2, 0x88>>2, 0x10>>2, 0x20>>2, 0x20>>2, 0x00>>2, 0x20>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x70>>2, 0x88>>2, 0x98>>2, 0xA8>>2, 0xB0>>2, 0x80>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x20>>2, 0x50>>2, 0x88>>2, 0x88>>2, 0xF8>>2, 0x88>>2, 0x88>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0xF0>>2, 0x48>>2, 0x48>>2, 0x70>>2, 0x48>>2, 0x48>>2, 0xF0>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x70>>2, 0x88>>2, 0x80>>2, 0x80>>2, 0x80>>2, 0x88>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0xF0>>2, 0x48>>2, 0x48>>2, 0x48>>2, 0x48>>2, 0x48>>2, 0xF0>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0xF8>>2, 0x80>>2, 0x80>>2, 0xF0>>2, 0x80>>2, 0x80>>2, 0xF8>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0xF8>>2, 0x80>>2, 0x80>>2, 0xF0>>2, 0x80>>2, 0x80>>2, 0x80>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x70>>2, 0x88>>2, 0x80>>2, 0x80>>2, 0x98>>2, 0x88>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0xF8>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x70>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x38>>2, 0x10>>2, 0x10>>2, 0x10>>2, 0x10>>2, 0x90>>2, 0x60>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x88>>2, 0x90>>2, 0xA0>>2, 0xC0>>2, 0xA0>>2, 0x90>>2, 0x88>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x80>>2, 0x80>>2, 0x80>>2, 0x80>>2, 0x80>>2, 0x80>>2, 0xF8>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x88>>2, 0x88>>2, 0xD8>>2, 0xA8>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x88>>2, 0x88>>2, 0xC8>>2, 0xA8>>2, 0x98>>2, 0x88>>2, 0x88>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x70>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0xF0>>2, 0x88>>2, 0x88>>2, 0xF0>>2, 0x80>>2, 0x80>>2, 0x80>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x70>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0xA8>>2, 0x70>>2, 0x08>>2, 0x00>>2, },
{ 0x00>>2, 0xF0>>2, 0x88>>2, 0x88>>2, 0xF0>>2, 0xA0>>2, 0x90>>2, 0x88>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x70>>2, 0x88>>2, 0x80>>2, 0x70>>2, 0x08>>2, 0x88>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0xF8>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x50>>2, 0x50>>2, 0x50>>2, 0x20>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0xA8>>2, 0xA8>>2, 0xD8>>2, 0x88>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x88>>2, 0x88>>2, 0x50>>2, 0x20>>2, 0x50>>2, 0x88>>2, 0x88>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x88>>2, 0x88>>2, 0x50>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0xF8>>2, 0x08>>2, 0x10>>2, 0x20>>2, 0x40>>2, 0x80>>2, 0xF8>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x70>>2, 0x40>>2, 0x40>>2, 0x40>>2, 0x40>>2, 0x40>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x80>>2, 0x80>>2, 0x40>>2, 0x20>>2, 0x10>>2, 0x08>>2, 0x08>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x70>>2, 0x10>>2, 0x10>>2, 0x10>>2, 0x10>>2, 0x10>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x20>>2, 0x50>>2, 0x88>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0xF8>>2, 0x00>>2, },
{ 0x20>>2, 0x10>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x70>>2, 0x08>>2, 0x78>>2, 0x88>>2, 0x78>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x80>>2, 0x80>>2, 0xB0>>2, 0xC8>>2, 0x88>>2, 0xC8>>2, 0xB0>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x70>>2, 0x88>>2, 0x80>>2, 0x88>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x08>>2, 0x08>>2, 0x68>>2, 0x98>>2, 0x88>>2, 0x98>>2, 0x68>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x70>>2, 0x88>>2, 0xF8>>2, 0x80>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x30>>2, 0x48>>2, 0x40>>2, 0xF0>>2, 0x40>>2, 0x40>>2, 0x40>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x78>>2, 0x88>>2, 0x88>>2, 0x78>>2, 0x08>>2, 0x88>>2, 0x70>>2, },
{ 0x00>>2, 0x80>>2, 0x80>>2, 0xB0>>2, 0xC8>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x20>>2, 0x00>>2, 0x60>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x08>>2, 0x00>>2, 0x18>>2, 0x08>>2, 0x08>>2, 0x08>>2, 0x48>>2, 0x48>>2, 0x30>>2, },
{ 0x00>>2, 0x80>>2, 0x80>>2, 0x88>>2, 0x90>>2, 0xE0>>2, 0x90>>2, 0x88>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x60>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0xD0>>2, 0xA8>>2, 0xA8>>2, 0xA8>>2, 0x88>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0xB0>>2, 0xC8>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x70>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0xB0>>2, 0xC8>>2, 0x88>>2, 0xC8>>2, 0xB0>>2, 0x80>>2, 0x80>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x68>>2, 0x98>>2, 0x88>>2, 0x98>>2, 0x68>>2, 0x08>>2, 0x08>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0xB0>>2, 0xC8>>2, 0x80>>2, 0x80>>2, 0x80>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x70>>2, 0x80>>2, 0x70>>2, 0x08>>2, 0xF0>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x40>>2, 0x40>>2, 0xF0>>2, 0x40>>2, 0x40>>2, 0x48>>2, 0x30>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x98>>2, 0x68>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x88>>2, 0x88>>2, 0x50>>2, 0x50>>2, 0x20>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x88>>2, 0x88>>2, 0xA8>>2, 0xA8>>2, 0x50>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x88>>2, 0x50>>2, 0x20>>2, 0x50>>2, 0x88>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x88>>2, 0x88>>2, 0x98>>2, 0x68>>2, 0x08>>2, 0x88>>2, 0x70>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0xF8>>2, 0x10>>2, 0x20>>2, 0x40>>2, 0xF8>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x18>>2, 0x20>>2, 0x10>>2, 0x60>>2, 0x10>>2, 0x20>>2, 0x18>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x60>>2, 0x10>>2, 0x20>>2, 0x18>>2, 0x20>>2, 0x10>>2, 0x60>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x48>>2, 0xA8>>2, 0x90>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x20>>2, 0x00>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x20>>2, 0x78>>2, 0xA0>>2, 0xA0>>2, 0xA0>>2, 0x78>>2, 0x20>>2, 0x00>>2, },
{ 0x00>>2, 0x30>>2, 0x48>>2, 0x40>>2, 0xE0>>2, 0x40>>2, 0x48>>2, 0xB0>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x88>>2, 0x70>>2, 0x50>>2, 0x70>>2, 0x88>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x88>>2, 0x88>>2, 0x50>>2, 0x20>>2, 0xF8>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x00>>2, },
{ 0x00>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x00>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x70>>2, 0x80>>2, 0xE0>>2, 0x90>>2, 0x48>>2, 0x38>>2, 0x08>>2, 0x70>>2, 0x00>>2, },
{ 0x50>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x70>>2, 0x88>>2, 0xA8>>2, 0xC8>>2, 0xA8>>2, 0x88>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x38>>2, 0x48>>2, 0x58>>2, 0x28>>2, 0x00>>2, 0x78>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x24>>2, 0x48>>2, 0x90>>2, 0x48>>2, 0x24>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x78>>2, 0x08>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x78>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x70>>2, 0x88>>2, 0xE8>>2, 0xC8>>2, 0xC8>>2, 0x88>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0xF8>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x20>>2, 0x50>>2, 0x20>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x20>>2, 0x20>>2, 0xF8>>2, 0x20>>2, 0x20>>2, 0xF8>>2, 0x00>>2, 0x00>>2, },
{ 0x30>>2, 0x48>>2, 0x10>>2, 0x20>>2, 0x78>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x70>>2, 0x08>>2, 0x30>>2, 0x08>>2, 0x70>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x10>>2, 0x20>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0xC8>>2, 0xB0>>2, 0x80>>2, 0x00>>2, },
{ 0x00>>2, 0x78>>2, 0xE8>>2, 0xE8>>2, 0x68>>2, 0x28>>2, 0x28>>2, 0x28>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x20>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x10>>2, 0x20>>2, },
{ 0x20>>2, 0x60>>2, 0x20>>2, 0x20>>2, 0x70>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x30>>2, 0x48>>2, 0x48>>2, 0x30>>2, 0x00>>2, 0x78>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x90>>2, 0x48>>2, 0x24>>2, 0x48>>2, 0x90>>2, 0x00>>2, 0x00>>2, },
{ 0x40>>2, 0xC0>>2, 0x40>>2, 0x40>>2, 0xE4>>2, 0x0C>>2, 0x14>>2, 0x3C>>2, 0x04>>2, 0x00>>2, },
{ 0x40>>2, 0xC0>>2, 0x40>>2, 0x40>>2, 0xE8>>2, 0x14>>2, 0x04>>2, 0x08>>2, 0x1C>>2, 0x00>>2, },
{ 0xC0>>2, 0x20>>2, 0x40>>2, 0x20>>2, 0xC8>>2, 0x18>>2, 0x28>>2, 0x78>>2, 0x08>>2, 0x00>>2, },
{ 0x00>>2, 0x20>>2, 0x00>>2, 0x20>>2, 0x20>>2, 0x40>>2, 0x88>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x40>>2, 0x20>>2, 0x70>>2, 0x88>>2, 0x88>>2, 0xF8>>2, 0x88>>2, 0x88>>2, 0x00>>2, 0x00>>2, },
{ 0x10>>2, 0x20>>2, 0x70>>2, 0x88>>2, 0x88>>2, 0xF8>>2, 0x88>>2, 0x88>>2, 0x00>>2, 0x00>>2, },
{ 0x20>>2, 0x50>>2, 0x70>>2, 0x88>>2, 0x88>>2, 0xF8>>2, 0x88>>2, 0x88>>2, 0x00>>2, 0x00>>2, },
{ 0x48>>2, 0xB0>>2, 0x70>>2, 0x88>>2, 0x88>>2, 0xF8>>2, 0x88>>2, 0x88>>2, 0x00>>2, 0x00>>2, },
{ 0x50>>2, 0x00>>2, 0x70>>2, 0x88>>2, 0x88>>2, 0xF8>>2, 0x88>>2, 0x88>>2, 0x00>>2, 0x00>>2, },
{ 0x20>>2, 0x50>>2, 0x70>>2, 0x88>>2, 0x88>>2, 0xF8>>2, 0x88>>2, 0x88>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x3C>>2, 0x50>>2, 0x90>>2, 0x9C>>2, 0xF0>>2, 0x90>>2, 0x9C>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x70>>2, 0x88>>2, 0x80>>2, 0x80>>2, 0x80>>2, 0x88>>2, 0x70>>2, 0x20>>2, 0x40>>2, },
{ 0x40>>2, 0xF8>>2, 0x80>>2, 0x80>>2, 0xF0>>2, 0x80>>2, 0x80>>2, 0xF8>>2, 0x00>>2, 0x00>>2, },
{ 0x10>>2, 0xF8>>2, 0x80>>2, 0x80>>2, 0xF0>>2, 0x80>>2, 0x80>>2, 0xF8>>2, 0x00>>2, 0x00>>2, },
{ 0x20>>2, 0xF8>>2, 0x80>>2, 0x80>>2, 0xF0>>2, 0x80>>2, 0x80>>2, 0xF8>>2, 0x00>>2, 0x00>>2, },
{ 0x50>>2, 0xF8>>2, 0x80>>2, 0x80>>2, 0xF0>>2, 0x80>>2, 0x80>>2, 0xF8>>2, 0x00>>2, 0x00>>2, },
{ 0x40>>2, 0x20>>2, 0x70>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x10>>2, 0x20>>2, 0x70>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x20>>2, 0x50>>2, 0x70>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x50>>2, 0x00>>2, 0x70>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0xF0>>2, 0x48>>2, 0x48>>2, 0xE8>>2, 0x48>>2, 0x48>>2, 0xF0>>2, 0x00>>2, 0x00>>2, },
{ 0x28>>2, 0x50>>2, 0x88>>2, 0xC8>>2, 0xA8>>2, 0x98>>2, 0x88>>2, 0x88>>2, 0x00>>2, 0x00>>2, },
{ 0x40>>2, 0x20>>2, 0x70>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x10>>2, 0x20>>2, 0x70>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x20>>2, 0x50>>2, 0x70>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x28>>2, 0x50>>2, 0x70>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x50>>2, 0x00>>2, 0x70>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x88>>2, 0x50>>2, 0x20>>2, 0x50>>2, 0x88>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x70>>2, 0x98>>2, 0x98>>2, 0xA8>>2, 0xC8>>2, 0xC8>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x40>>2, 0x20>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x10>>2, 0x20>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x20>>2, 0x50>>2, 0x00>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x50>>2, 0x00>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x10>>2, 0x20>>2, 0x88>>2, 0x88>>2, 0x50>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x80>>2, 0xF0>>2, 0x88>>2, 0xF0>>2, 0x80>>2, 0x80>>2, 0x80>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x70>>2, 0x88>>2, 0x90>>2, 0xA0>>2, 0x90>>2, 0x88>>2, 0xB0>>2, 0x00>>2, 0x00>>2, },
{ 0x40>>2, 0x20>>2, 0x00>>2, 0x70>>2, 0x08>>2, 0x78>>2, 0x88>>2, 0x78>>2, 0x00>>2, 0x00>>2, },
{ 0x10>>2, 0x20>>2, 0x00>>2, 0x70>>2, 0x08>>2, 0x78>>2, 0x88>>2, 0x78>>2, 0x00>>2, 0x00>>2, },
{ 0x20>>2, 0x50>>2, 0x00>>2, 0x70>>2, 0x08>>2, 0x78>>2, 0x88>>2, 0x78>>2, 0x00>>2, 0x00>>2, },
{ 0x28>>2, 0x50>>2, 0x00>>2, 0x70>>2, 0x08>>2, 0x78>>2, 0x88>>2, 0x78>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x50>>2, 0x00>>2, 0x70>>2, 0x08>>2, 0x78>>2, 0x88>>2, 0x78>>2, 0x00>>2, 0x00>>2, },
{ 0x20>>2, 0x50>>2, 0x20>>2, 0x70>>2, 0x08>>2, 0x78>>2, 0x88>>2, 0x78>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x78>>2, 0x14>>2, 0x7C>>2, 0x90>>2, 0x7C>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x70>>2, 0x88>>2, 0x80>>2, 0x88>>2, 0x70>>2, 0x20>>2, 0x40>>2, },
{ 0x40>>2, 0x20>>2, 0x00>>2, 0x70>>2, 0x88>>2, 0xF8>>2, 0x80>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x10>>2, 0x20>>2, 0x00>>2, 0x70>>2, 0x88>>2, 0xF8>>2, 0x80>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x20>>2, 0x50>>2, 0x00>>2, 0x70>>2, 0x88>>2, 0xF8>>2, 0x80>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x50>>2, 0x00>>2, 0x70>>2, 0x88>>2, 0xF8>>2, 0x80>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x40>>2, 0x20>>2, 0x00>>2, 0x60>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x20>>2, 0x40>>2, 0x00>>2, 0x60>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x20>>2, 0x50>>2, 0x00>>2, 0x60>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x50>>2, 0x00>>2, 0x60>>2, 0x20>>2, 0x20>>2, 0x20>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0xC0>>2, 0x30>>2, 0x70>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x28>>2, 0x50>>2, 0x00>>2, 0xB0>>2, 0xC8>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x00>>2, 0x00>>2, },
{ 0x40>>2, 0x20>>2, 0x00>>2, 0x70>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x10>>2, 0x20>>2, 0x00>>2, 0x70>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x20>>2, 0x50>>2, 0x00>>2, 0x70>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x28>>2, 0x50>>2, 0x00>>2, 0x70>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x50>>2, 0x00>>2, 0x70>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x70>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x20>>2, 0x00>>2, 0xF8>>2, 0x00>>2, 0x20>>2, 0x00>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x00>>2, 0x00>>2, 0x78>>2, 0x98>>2, 0xA8>>2, 0xC8>>2, 0xF0>>2, 0x00>>2, 0x00>>2, },
{ 0x40>>2, 0x20>>2, 0x00>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x98>>2, 0x68>>2, 0x00>>2, 0x00>>2, },
{ 0x10>>2, 0x20>>2, 0x00>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x98>>2, 0x68>>2, 0x00>>2, 0x00>>2, },
{ 0x20>>2, 0x50>>2, 0x00>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x98>>2, 0x68>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x50>>2, 0x00>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0x98>>2, 0x68>>2, 0x00>>2, 0x00>>2, },
{ 0x00>>2, 0x10>>2, 0x20>>2, 0x88>>2, 0x88>>2, 0x98>>2, 0x68>>2, 0x08>>2, 0x88>>2, 0x70>>2, },
{ 0x00>>2, 0x00>>2, 0x80>>2, 0xF0>>2, 0x88>>2, 0x88>>2, 0x88>>2, 0xF0>>2, 0x80>>2, 0x80>>2, },
{ 0x00>>2, 0x50>>2, 0x00>>2, 0x88>>2, 0x88>>2, 0x98>>2, 0x68>>2, 0x08>>2, 0x88>>2, 0x70>>2, },
};

static sdl_font sdl_default_font;

void (*sdl_printfchar)(sdl_font *f, unsigned char c);

void sdl_printfchar15(sdl_font *f, unsigned char c)
{
  unsigned short *dst=&((unsigned short*)_hw_screen->pixels)[f->x+f->y*(_hw_screen->pitch>>1)],w,h=f->h;
//unsigned char  *src=f->data[ (c%16)*f->w + (c/16)*f->h ];
  unsigned char  *src=&f->data[c*10];

 if(f->solid)
         while(h--)
         {
          w=f->wmask;
          while(w)
          {
           if( *src & w ) *dst++=f->fg; else *dst++=f->bg;
           w>>=1;
          }
          src++;    

          dst+=(_hw_screen->pitch>>1)-(f->w);
         }
 else
         while(h--)
         {
          w=f->wmask;
          while(w)
          {
           if( *src & w ) *dst=f->fg;
           dst++;
           w>>=1;
          }
          src++;

          dst+=(_hw_screen->pitch>>1)-(f->w);
         }
}

void sdl_printf(sdl_font *f, int x, int y, const char *format, ...)
{
 char buffer[4096]; int c; sdl_font *g=&sdl_default_font;
 va_list  args;

 va_start(args, format);
 vsprintf(buffer, format, args);
 va_end(args);

 if(f!=NULL) g=f;

 if(x<0) x=g->x; else g->x=x; 
 if(y<0) y=g->y; else g->y=y;

 for(c=0;buffer[c];c++)
 {
  switch(buffer[c])
  {
   case '\b': g->x=x;g->y=y; break;

   case '\n': g->y+=g->h;
   case '\r': g->x=x;
              break;

   default:   sdl_printfchar(g, (unsigned char)buffer[c]);
              g->x+=g->w;
              break;
  }
 }
}

void sdl_printf_init(sdl_font *f, int w, int h, void *data, int fg, int bg, int solid)
{
 sdl_printfchar=sdl_printfchar15;
 f->x=f->y=0;
 f->wmask=1<<(w-1);
 f->w=w;
 f->h=h;
 f->data=(unsigned char *)data;
 f->fg=fg;
 f->bg=bg;
 f->solid=solid;
}

static char _sip_temp[30];
static char _sip_text[30];
static unsigned int _sip_cur_char = 0;

static char _sip_lookup[7][20] = 
{
	{'1','2','3','4','5','6','7','8','9','0','-','+','=',  0,  0,	0,0,0,0,0},
	{'Q','W','E','R','T','Y','U','I','O','P','{','}',  0,  0,  0,   0,0,0,0,0},
	{'A','S','D','F','G','H','J','K','L',':',';','[',']',  0,  0,   0,0,0,0,0},
	{'Z','X','C','V','B','N','M','<',',','.','>','/','?',  0,  0,	0,0,0,0,0},
	{'`','!','@','#','$','%','^','&','*','(',')',  0,  0,  0,  0,	0,0,0,0,0},
	{'C','C','A','A','D','D','S','S','E','E',  0,  0,  0,  0,  0,	0,0,0,0,0},
	{' ',' ','T','T','P','P','I','I','B','B',  0,  0,  0,  0,  0,	0,0,0,0,0}
};


static void SdlVideoDrawSIP(void)
{
	memset(_hw_screen->pixels, 0, _hw_screen->pitch * _hw_screen->h);

	sdl_printf(NULL, 0, 0,  "1 2 3 4 5 6 7 8 9 0 - + =");
	sdl_printf(NULL, 0, 20, "Q W E R T Y U I O P { }");
	sdl_printf(NULL, 0, 40, "A S D F G H J K L : ; [ ]");
	sdl_printf(NULL, 0, 60, "Z X C V B N M < , . > / ?");
	sdl_printf(NULL, 0, 80, "` ~ ! @ # $ % ^ & * ( )");
	sdl_printf(NULL, 0, 100, "CTL ALT BKS SHF ENT");
	sdl_printf(NULL, 0, 120, "SPC TAB ESC INS <<<");
	sdl_printf(NULL, 0, 140, _sip_text);
}

static void SdlVideoMakeDirty(int left, int top, int width, int height)
{
	if (_num_dirty_rects < MAX_DIRTY_RECTS) {
			_dirty_rects[_num_dirty_rects].x = left;
			_dirty_rects[_num_dirty_rects].y = top;
			_dirty_rects[_num_dirty_rects].w = width;
			_dirty_rects[_num_dirty_rects].h = height;
	}
	_num_dirty_rects++;
}

static void UpdatePalette(uint start, uint count)
{
	SDL_Color pal[256];
	uint i;

	for (i = 0; i != count; i++) {
		pal[i].r = _cur_palette[start + i].r;
		pal[i].g = _cur_palette[start + i].g;
		pal[i].b = _cur_palette[start + i].b;
		pal[i].unused = 0;
	}

	SDL_SetColors(_hw_screen, pal, start, count);

#if 0
	if( _textinput_enabled == false )
	{
		SDL_SetColors(_hw_screen, pal, start, count);
	}
	else
	{
		SipShowIM(SIPF_OFF);
		SDL_SetColors(_hw_screen, pal, start, count);
		SipShowIM(SIPF_ON);
	}
#endif
}

static void InitPalette(void)
{
	UpdatePalette(0, 256);
}

static void CheckPaletteAnim(void)
{
	if (_pal_last_dirty != -1) {
			UpdatePalette(_pal_first_dirty, _pal_last_dirty - _pal_first_dirty + 1);
			_pal_last_dirty = -1;
	}
}

static void DrawSurfaceToScreen(void)
{
	if( _textinput_enabled == true )
	{
		SdlVideoDrawSIP();
		SDL_CALL SDL_UpdateRect(_hw_screen, 0, 0, 0, 0);
		return;
	}

	if( _num_dirty_rects == 0 )
	{
		return;
	}

	if( _stretched != true )
	{
		SDL_Rect src_rect, dst_rect;

		dst_rect.x = 0;
		dst_rect.y = 0;
		dst_rect.w = _hw_screen->w;
		dst_rect.h = _hw_screen->h;

		src_rect = dst_rect;

		SDL_LowerBlit(_sw_screen, &src_rect, _hw_screen, &dst_rect);
	}
	else
	{
		SDL_SoftStretch(	_sw_screen, NULL,
							_hw_screen, NULL);

		SDL_UpdateRect(_hw_screen, 0, 0, 0, 0);
		_num_dirty_rects = 0;
		return;
	}

	_num_dirty_rects = 0;
}

static void QueryResolution(unsigned int *res_x, unsigned int *res_y)
{
	SDL_Rect rect;
	SDL_GP2X_GetPhysicalScreenSize(&rect);
	*res_x = rect.w;
	*res_y = rect.h;
}

bool CreateMainSurface(void)
{
	extern const char _openttd_revision[];
	SDL_Surface *newscreen;
	char caption[50];
	
	QueryResolution(&_width, &_height);
	if( _height > _width ) 
	{
		unsigned int temp;
		_rotated = true;
		temp = _width;
		_width = _height;
		_height = temp;
	} else {
		_rotated = false;	
	}

	_cur_resolution[0] = (_width < 320 ? 320 : (_width > 640 ? 640 : _width)); ///10*11; //480;
	_cur_resolution[1] = (_height < 240 ? 240 : (_height > 480 ? 480 : _height)); ///10*11; //272;

	if( !(_width == _cur_resolution[0] && _height == _cur_resolution[1]) ) 
	{
		_stretched = true;
	}
	else 
	{
		_stretched = false;
	}

	SDL_CALL SDL_GP2X_AllowGfxMemory(NULL, 0);

	newscreen = SDL_CALL SDL_SetVideoMode(_width, _height, 8, SDL_HWSURFACE | SDL_HWPALETTE | SDL_FULLSCREEN);
	if (newscreen == NULL)
		return false;

	_hw_screen = newscreen;

	newscreen = SDL_CALL SDL_CreateRGBSurface(SDL_HWSURFACE, _cur_resolution[0], _cur_resolution[1], 8, 0, 0, 0, 0);
	if (newscreen == NULL)
		return false;
	
	if( _stretched == true )
	{
		zoom_x = (float)_cur_resolution[0] / (float)_width;
		zoom_y = (float)_cur_resolution[1] / (float)_height;
	}

	_screen.width = _cur_resolution[0];
	_screen.height = _cur_resolution[1];
	_screen.pitch = _cur_resolution[0];

	_sw_screen = newscreen;

	InitPalette();

	//snprintf(caption, sizeof(caption), "OpenTTD %s", _openttd_revision);
	//SDL_CALL SDL_WM_SetCaption(caption, caption);
	SDL_CALL SDL_ShowCursor(0);

	GameSizeChanged();


	//DKS - new touchscreen stuff, detect if we are on an F200 or not:
	if (SDL_GP2X_MouseType() == 2) {
		//we are on an F200, there is a touchscreen.
		_f200_controls = !_want_F100;  // did user request F100 style controls via command line?
		_right_handed = !_want_left_handed;	// did user request left-handed stylus controls via command-line
		_ext_mouse = false;	
	} else {
		if (SDL_GP2X_MouseType() == 1) {
			// external mouse detected
			_f200_controls = false;
			_ext_mouse = true;
		} else {
			// we must be on an F100, and no external mouse, either
			_f200_controls = false;	// did user request left handed via command line?
			_ext_mouse = false;
		}
	}
		
	return true;
}

typedef struct VkMapping {
	uint16 vk_from;
	byte vk_count;
	byte map_to;
} VkMapping;

#define AS(x, z) {x, 0, z}
#define AM(x, y, z, w) {x, y - x, z}

static const VkMapping _vk_mapping[] = {
	// Pageup stuff + up/down
	AM(SDLK_PAGEUP, SDLK_PAGEDOWN, WKC_PAGEUP, WKC_PAGEDOWN),
	AS(SDLK_UP,     WKC_UP),
	AS(SDLK_DOWN,   WKC_DOWN),
	AS(SDLK_LEFT,   WKC_LEFT),
	AS(SDLK_RIGHT,  WKC_RIGHT),

	AS(SDLK_HOME,   WKC_HOME),
	AS(SDLK_END,    WKC_END),

	AS(SDLK_INSERT, WKC_INSERT),
	AS(SDLK_DELETE, WKC_DELETE),

	// Map letters & digits
	AM(65, 90, 'a', 'z'),
	AM(SDLK_a, SDLK_z, 'A', 'Z'),
	AM(SDLK_0, SDLK_9, '0', '9'),

	AS(SDLK_ESCAPE,    WKC_ESC),
	AS(SDLK_PAUSE,     WKC_PAUSE),
	AS(SDLK_BACKSPACE, WKC_BACKSPACE),

	AS(SDLK_SPACE,     WKC_SPACE),
	AS(SDLK_RETURN,    WKC_RETURN),
	AS(SDLK_TAB,       WKC_TAB),

	// Function keys
	AM(SDLK_F1, SDLK_F12, WKC_F1, WKC_F12),

	// Numeric part.
	// What is the virtual keycode for numeric enter??
	AM(SDLK_KP0, SDLK_KP9, WKC_NUM_0, WKC_NUM_9),
	AS(SDLK_KP_DIVIDE,   WKC_NUM_DIV),
	AS(SDLK_KP_MULTIPLY, WKC_NUM_MUL),
	AS(SDLK_KP_MINUS,    WKC_NUM_MINUS),
	AS(SDLK_KP_PLUS,     WKC_NUM_PLUS),
	AS(SDLK_KP_ENTER,    WKC_NUM_ENTER),
	AS(SDLK_KP_PERIOD,   WKC_NUM_DECIMAL)
};

static uint32 ConvertSdlKeyIntoMy(SDL_keysym *sym)
{
	const VkMapping *map;
	uint key = 0;

	for (map = _vk_mapping; map != endof(_vk_mapping); ++map) {
		if ((uint)(sym->sym - map->vk_from) <= map->vk_count) {
			key = sym->sym - map->vk_from + map->map_to;
			break;
		}
	}

	// check scancode for BACKQUOTE key, because we want the key left of "1", not anything else (on non-US keyboards)
#if defined(WIN32) || defined(__OS2__)
	if (sym->scancode == 41) key = WKC_BACKQUOTE;
#elif defined(__APPLE__)
	if (sym->scancode == 10) key = WKC_BACKQUOTE;
#elif defined(__MORPHOS__)
	if (sym->scancode == 0)  key = WKC_BACKQUOTE;  // yes, that key is code '0' under MorphOS :)
#elif defined(__BEOS__)
	if (sym->scancode == 17) key = WKC_BACKQUOTE;
#elif defined(__SVR4) && defined(__sun)
	if (sym->scancode == 60) key = WKC_BACKQUOTE;
	if (sym->scancode == 49) key = WKC_BACKSPACE;
#elif defined(__sgi__)
	if (sym->scancode == 22) key = WKC_BACKQUOTE;
#else
	if (sym->scancode == 49) key = WKC_BACKQUOTE;
#endif

	// META are the command keys on mac
	if (sym->mod & KMOD_META)  key |= WKC_META;
	if (sym->mod & KMOD_SHIFT) key |= WKC_SHIFT;
	if (sym->mod & KMOD_CTRL)  key |= WKC_CTRL;
	if (sym->mod & KMOD_ALT)   key |= WKC_ALT;
	// these two lines really help porting hotkey combos. Uncomment to use -- Bjarni
#if 0
	DEBUG(driver, 0) ("scancode character pressed %u", sym->scancode);
	DEBUG(driver, 0) ("unicode character pressed %u", sym->unicode);
#endif
	if( key == 0 ) return 0;

	return (key << 16) + sym->unicode;
}
static int PollEvent(void)
{
	SDL_Event ev;
	float _x, _y;

	if (!SDL_CALL SDL_PollEvent(&ev)) return -2;

	switch (ev.type) {
		case SDL_MOUSEMOTION:
			
			//DKS - old TTD2X stuff:
			/*
			if( _textinput_enabled == true )
			{
				break;
			}

			_x = ev.motion.x;
			_y = ev.motion.y;

			if( _stretched )
			{
				_x *= zoom_x;
				_y *= zoom_y;
			}

			if (_cursor.fix_at) {
				int dx = _x - _cursor.pos.x;
				int dy = _y - _cursor.pos.y;
				if (dx != 0 || dy != 0) {
					_cursor.delta.x += dx;
					_cursor.delta.y += dy;
					//SDL_CALL SDL_WarpMouse(_cursor.pos.x, _cursor.pos.y);
				}
			} else {
				_cursor.delta.x = _x - _cursor.pos.x;
				_cursor.delta.y = _y - _cursor.pos.y;
				_cursor.pos.x = _x;
				_cursor.pos.y = _y;
				_cursor.dirty = true;
			}

			_left_button_down = true;
			*/	
		
			_x = ev.motion.x;
			_y = ev.motion.y;

			if( _stretched )
			{
				_x *= zoom_x;
				_y *= zoom_y;
			}

			if (_cursor.fix_at) {
				int dx = _x - _cursor.pos.x;
				int dy = _y - _cursor.pos.y;
				if (dx != 0 || dy != 0) {
					_cursor.delta.x += dx;
					_cursor.delta.y += dy;
					//SDL_CALL SDL_WarpMouse(_cursor.pos.x, _cursor.pos.y);
				}
			} else {
				_cursor.delta.x = _x - _cursor.pos.x;
				_cursor.delta.y = _y - _cursor.pos.y;
				_cursor.pos.x = _x;
				_cursor.pos.y = _y;
				_cursor.dirty = true;
			}
			break;
			

		case SDL_MOUSEBUTTONDOWN:
			//DKS - old TTD2X stuff:
			/*
			if( _textinput_enabled == true )
			{
				break;
			}
			_x = ev.button.x;
			_y = ev.button.y;

			if( _stretched )
			{
				_x *= zoom_x;
				_y *= zoom_y;
			}

			if (_cursor.fix_at) {
				int dx = _x - _cursor.pos.x;
				int dy = _y - _cursor.pos.y;
				if (dx != 0 || dy != 0) {
					_cursor.delta.x += dx;
					_cursor.delta.y += dy;
					//SDL_CALL SDL_WarpMouse(_cursor.pos.x, _cursor.pos.y);
				}
			} else {
				_cursor.delta.x = _x - _cursor.pos.x;
				_cursor.delta.y = _y - _cursor.pos.y;
				_cursor.pos.x = _x;
				_cursor.pos.y = _y;
				_cursor.dirty = true;
			}

			switch (ev.button.button) {
				case SDL_BUTTON_LEFT:
					_left_button_down = true;
					break;

				case SDL_BUTTON_RIGHT:
					_right_button_down = true;
					_right_button_clicked = true;
					break;

				case SDL_BUTTON_WHEELUP:   _cursor.wheel--; break;
				case SDL_BUTTON_WHEELDOWN: _cursor.wheel++; break;

				default: break;
			}
			break;
			*/

			if (_ext_mouse) {	
				if (_rightclick_emulate && SDL_CALL SDL_GetModState() & KMOD_CTRL) {
					ev.button.button = SDL_BUTTON_RIGHT;
				}

				switch (ev.button.button) {
					case SDL_BUTTON_LEFT:
						_left_button_down = true;
						break;
	
					case SDL_BUTTON_RIGHT:
						_right_button_down = true;
						_right_button_clicked = true;
						break;

					case SDL_BUTTON_WHEELUP:   _cursor.wheel--; break;
					case SDL_BUTTON_WHEELDOWN: _cursor.wheel++; break;

					default: break;
				}
				HandleMouseEvents();
			}
			

			break;

		case SDL_MOUSEBUTTONUP:
			//DKS - old TTD2X stuff:
			/*
			if( _textinput_enabled == true )
			{
				unsigned int btn_x = ev.button.x / 24;
				unsigned int btn_y = ev.button.y / 20;
				
				if( (btn_y == 5 || btn_y == 6)			||
					((ev.button.x / 12) % 2 == 0	&&
					 (ev.button.y / 10) % 2 == 0	&&
					 btn_y < 7 && btn_x < 20		)	)
				{
					if( btn_y == 5 || btn_y == 6 )
					{
						switch(_sip_lookup[btn_y][btn_x])
						{
						case 'B':
							if( _sip_cur_char > 0 )
							{
								_sip_text[_sip_cur_char] = 0;
								_sip_cur_char--;
							}
							HandleKeypress((WKC_BACKSPACE << 16));
							break;
						case 'D':
							if( _sip_cur_char > 0 )
							{
								_sip_text[_sip_cur_char] = 0;
								_sip_cur_char--;
							}
							HandleKeypress((WKC_BACKSPACE << 16));
							break;
						case ' ':
							if( _sip_cur_char >= 29 )
							{
								unsigned int i;
								for(i = 0; i < _sip_cur_char; i++)
								{
									_sip_text[i] = _sip_text[i + 1];
								}
								_sip_text[_sip_cur_char] = 0;
							}
							_sip_text[_sip_cur_char] = _sip_lookup[btn_y][btn_x];
							HandleKeypress((unsigned long)(WKC_SPACE << 16) + _sip_text[_sip_cur_char]);
							if( _sip_cur_char < 29 )
							{
								_sip_cur_char++;
							}
							break;
						case 'T':
							HandleKeypress((WKC_TAB << 16));
							break;
						case 'P':
							HandleKeypress((WKC_ESC << 16));
							break;
						case 'I':
							HandleKeypress((WKC_INSERT << 16));
							break;
						}
					}
					else
					{
						if( (int)_sip_lookup[btn_y][btn_x] != 0 )
						{
							if( _sip_cur_char >= 29 )
							{
								unsigned int i;
								for(i = 0; i < _sip_cur_char; i++)
								{
									_sip_text[i] = _sip_text[i + 1];
								}
								_sip_text[_sip_cur_char] = 0;
							}
							
							_sip_text[_sip_cur_char] = _sip_lookup[btn_y][btn_x];
							HandleKeypress((unsigned long)_sip_text[_sip_cur_char]);

							if( _sip_cur_char < 29 )
							{
								_sip_cur_char++;
							}
						}
					}
				}
			}
			else
			{
				_x = ev.button.x;
				_y = ev.button.y;

				if( _stretched )
				{
					_x *= zoom_x;
					_y *= zoom_y;
				}

				if (_cursor.fix_at) {
					int dx = _x - _cursor.pos.x;
					int dy = _y - _cursor.pos.y;
					if (dx != 0 || dy != 0) {
						_cursor.delta.x += dx;
						_cursor.delta.y += dy;
						//SDL_CALL SDL_WarpMouse(_cursor.pos.x, _cursor.pos.y);
					}
				} else {
					_cursor.delta.x = _x - _cursor.pos.x;
					_cursor.delta.y = _y - _cursor.pos.y;
					_cursor.pos.x = _x;
					_cursor.pos.y = _y;
					_cursor.dirty = true;
				}

				if (_rightclick_emulate) {
					_right_button_down = false;
					_left_button_down = false;
					_left_button_clicked = false;
				} else if (ev.button.button == SDL_BUTTON_LEFT) {
					_left_button_down = false;
					_left_button_clicked = false;
				} else if (ev.button.button == SDL_BUTTON_RIGHT) {
					_right_button_down = false;
				}
			}
			break;
			*/
			if (_ext_mouse) {
				if (_rightclick_emulate) {
					_right_button_down = false;
					_left_button_down = false;
					_left_button_clicked = false;
				} else if (ev.button.button == SDL_BUTTON_LEFT) {
					_left_button_down = false;
					_left_button_clicked = false;
				} else if (ev.button.button == SDL_BUTTON_RIGHT) {
					_right_button_down = false;
				}
				HandleMouseEvents();
			}
			break;


		case SDL_ACTIVEEVENT:
			if (!(ev.active.state & SDL_APPMOUSEFOCUS)) break;

			if (ev.active.gain) { // mouse entered the window, enable cursor
				_cursor.in_window = true;
			}
			break;

		case SDL_QUIT: HandleExitGameRequest(); break;

		case SDL_KEYDOWN: /* Toggle full-screen on ALT + ENTER/F */
			HandleKeypress(ConvertSdlKeyIntoMy(&ev.key.keysym));
			break;
		
		case SDL_KEYUP:
#if 0
			if( _textinput_enabled == true )
			{
				_force_screenupdate = true;
			}
#endif
			break;

		case SDL_VIDEORESIZE: {
			int w = clamp(ev.resize.w, 64, MAX_SCREEN_WIDTH);
			int h = clamp(ev.resize.h, 64, MAX_SCREEN_HEIGHT);
			ChangeResInGame(w, h);
			break;
		}
	}
	return -1;
}

static const char *SdlVideoStart(const char * const *parm)
{
	char buf[30];

//DKS - had to add SDL_INIT_EVENTTHREAD for touchscreen
//	const char *s = SdlOpen(SDL_INIT_VIDEO|SDL_INIT_JOYSTICK);
	const char *s = SdlOpen(SDL_INIT_VIDEO|SDL_INIT_JOYSTICK|SDL_INIT_EVENTTHREAD);
	if (s != NULL) return s;

	SDL_CALL SDL_VideoDriverName(buf, 30);
	DEBUG(driver, 1) ("sdl: using driver '%s'", buf);

	_rightclick_emulate = true;

	CreateMainSurface();

	MarkWholeScreenDirty();

	joystick = SDL_CALL SDL_JoystickOpen(0);
	SDL_CALL SDL_JoystickEventState(SDL_DISABLE);
	SDL_CALL SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);
	SDL_CALL SDL_EnableUNICODE(1);

	//init font
	sdl_printf_init(&sdl_default_font,6,10,sdl_fontf,0xFFFF,0x0000,1);

	return NULL;
}

static void SdlVideoStop(void)
{
	//gxFunc.GXCloseInput();
	SdlClose(SDL_INIT_VIDEO);
}

static void SdlVideoMainLoop(void)
{
	uint32 cur_ticks = SDL_CALL SDL_GetTicks();
	uint32 next_tick = cur_ticks + 30;
	uint32 pal_tick = 0;
	uint32 mod;
	int numkeys;
	Uint8 *keys;

	//DKS - new to allow us to prevent certain buttons used in hot-key combos from triggering normal events
	uint32 combo_pressed = 0;

	for (;;) {
		uint32 prev_cur_ticks = cur_ticks; // to check for wrapping
		InteractiveRandom(); // randomness

		while (PollEvent() == -1) {}
		if (_exit_game) return;

		mod = SDL_CALL SDL_GetModState();
		keys = SDL_CALL SDL_GetKeyState(&numkeys);

		/* Speedup when pressing tab, except when using ALT+TAB
		 * to switch to another application */
		if (keys[SDLK_TAB] && (mod & KMOD_ALT) == 0)
		{
			if (!_networking && _game_mode != GM_MENU) _fast_forward |= 2;
		} else if (_fast_forward & 2) {
			_fast_forward = 0;
		}

		cur_ticks = SDL_CALL SDL_GetTicks();
		if (cur_ticks >= next_tick || (_fast_forward && !_pause) || cur_ticks < prev_cur_ticks) {
			int ikey;
			next_tick = cur_ticks + 30;

			_ctrl_pressed  = !!(mod & KMOD_CTRL);
			_shift_pressed = !!(mod & KMOD_SHIFT);

			// determine which directional keys are down
#ifdef ARM
			SDL_CALL SDL_JoystickUpdate();

//DKS - disabling unless someone complains about new control scheme
			if (!_f200_controls) {
				// original F100-style GP2X controls:
				for(ikey = 0; ikey < 4; ikey++)
				{
					if( GetAsyncKeyState(ikey * 2) < 0 )
					{
						int _x, _y;
	
						_x = _cursor.pos.x;
						_y = _cursor.pos.y;
						switch(ikey * 2) // Cheap way of going through the dpad
						{
						case GP2X_UP:
							_y -= scrollrate;
							if( _y < 0 ) _y = 0;
							break;
						case GP2X_RIGHT:
							_x += scrollrate;
							if( _x > _cur_resolution[0] - 1 ) _x = _cur_resolution[0] - 1;
							break;
						case GP2X_DOWN:
							_y += scrollrate;
							if( _y > _cur_resolution[1] - 1 ) _y = _cur_resolution[1] - 1;
							break;
						case GP2X_LEFT:
							_x -= scrollrate;
							if( _x < 0 ) _x = 0;
							break;
						}
	
						if (_cursor.fix_at) 
						{
							int dx = _x - _cursor.pos.x;
							int dy = _y - _cursor.pos.y;
							if (dx != 0 || dy != 0) 
							{
								_cursor.delta.x += dx;
								_cursor.delta.y += dy;
							}
						}
						else 
						{
							_cursor.delta.x = _x - _cursor.pos.x;
							_cursor.delta.y = _y - _cursor.pos.y;
							_cursor.pos.x = _x;
							_cursor.pos.y = _y;
							_cursor.dirty = true;
						}
					}
				}
	

				//DKS - new zoom-in/out combos:
				if ((GetAsyncKeyState(GP2X_R) < 0) &&
					(GetAsyncKeyState(GP2X_Y) < 0))
				{
					// simulate Shift-F5 for zoom-in
					uint32 tmpkey = WKC_F5;
					tmpkey |= WKC_SHIFT;
					HandleKeypress(tmpkey << 16);
					SDL_Delay(250);
					combo_pressed = 1;
				} else if ((GetAsyncKeyState(GP2X_R) < 0) &&
					(GetAsyncKeyState(GP2X_X) < 0))
				{
					// simulate Shift-F6 for zoom-out
					uint32 tmpkey = WKC_F6;
					tmpkey |= WKC_SHIFT;
					HandleKeypress(tmpkey << 16);
					SDL_Delay(250);
					combo_pressed = 1;
				} else if ((GetAsyncKeyState(GP2X_R) < 0) &&
					(GetAsyncKeyState(GP2X_A) < 0))
				{
					// simulate 'a' for auto-rail hotkey
					uint32 tmpkey = 'A';
					HandleKeypress(tmpkey << 16);
					SDL_Delay(250);
					combo_pressed = 1;
				} else if ((GetAsyncKeyState(GP2X_R) < 0) &&
					(GetAsyncKeyState(GP2X_B) < 0))
				{
					// simulate 'x' for transparent buildings hotkey
					uint32 tmpkey = 'X';
					HandleKeypress(tmpkey << 16);
					SDL_Delay(250);
					combo_pressed = 1;
				} else if ((GetAsyncKeyState(GP2X_R) < 0) &&
					(GetAsyncKeyState(GP2X_SELECT) < 0))
				{
					// simulate 'l' for landscaping hotkey
					uint32 tmpkey = 'L';
					HandleKeypress(tmpkey << 16);
					SDL_Delay(250);
					combo_pressed = 1;
				} else {
					combo_pressed = 0;
				}

				//DKS - new hotkey, press all four scroll directions at once to bring up cheat menu
				if ((GetAsyncKeyState(GP2X_Y) < 0) && (GetAsyncKeyState(GP2X_X) < 0) 
					&& (GetAsyncKeyState(GP2X_A) < 0) && (GetAsyncKeyState(GP2X_B) < 0)) {
						// simulate 'Ctrl-Alt-c' for cheats hotkey
						uint32 tmpkey = 'C';
						tmpkey |= WKC_CTRL;
						tmpkey |= WKC_ALT;
						HandleKeypress(tmpkey << 16);
						SDL_Delay(250);
						combo_pressed = 1;
				}

				if (!combo_pressed) {
				
					/* Handle CONTROL button (that also might move window under cursor depending on context) */
					if (GetAsyncKeyState(GP2X_VOLDOWN) < 0)
					{
        				    Window *w;
				            w = FindWindowFromPt(_cursor.pos.x, _cursor.pos.y);

					        if (w->window_class != WC_MAIN_WINDOW &&
					                w->window_class != WC_SELECT_GAME &&
					                w->window_class != WC_MAIN_TOOLBAR &&
					                w->window_class != WC_STATUS_BAR &&
				        	        w->window_class != WC_TOOLBAR_MENU &&
				                	w->window_class != WC_TOOLTIPS) {
						                _left_button_down = true;
						                windowmove_pressed = true;
								//DKS - increase scroll rate to move these windows faster
								scrollrate = _fast_cursorrate;
				                } else {
		        		            _ctrl_pressed = true;
						}
					} else {
						if (windowmove_pressed) {
							windowmove_pressed = false;
							_left_button_down = false;
							_left_button_clicked = false;
							scrollrate = _slow_cursorrate;
						}
						_ctrl_pressed = false;
			                }
        

					if (( GetAsyncKeyState(GP2X_L) < 0) )
					{
						_left_button_down = true;
					}
					else
					{
						if (!windowmove_pressed) {
							_left_button_down = false;
							_left_button_clicked = false;
						}
					}
	
					if( GetAsyncKeyState(GP2X_R) < 0) 
					{
						_right_button_down = true;
						_right_button_clicked = _right_button_down;
					}
					else
					{
						_right_button_down = false;
					}	
					
					/* Delete windows */
					if (GetAsyncKeyState(GP2X_SELECT) < 0)
					{
						DeleteNonVitalWindows();
					}
				
					_dirkeys =	(GetAsyncKeyState(GP2X_A)	< 0	? 1 : 0) |
							(GetAsyncKeyState(GP2X_Y)	< 0	? 2 : 0) |
							(GetAsyncKeyState(GP2X_B)	< 0	? 4 : 0) |
							(GetAsyncKeyState(GP2X_X)	< 0	? 8 : 0);
					
				} else {
							_dirkeys = 0;
				}
					// don't want new hotkey combos to trigger scrolling
				


				/* speed up cursor when START is held down */
				if( GetAsyncKeyState(GP2X_START) < 0) 
				{
					scrollrate = _fast_cursorrate;
				} else {
					scrollrate = _slow_cursorrate;
				}

//				/* CONTROL pressed */
//				if (GetAsyncKeyState(GP2X_VOLDOWN) < 0)
//				{
//					_ctrl_pressed = true;
//				}
//				else
//				{
//					_ctrl_pressed = false;
//				}
	
				/* SHIFT pressed */
				if (GetAsyncKeyState(GP2X_VOLUP) < 0)
				{
					_shift_pressed = true;
				}
				else
				{
					_shift_pressed = false;
				}

			} //  end of !_f200_controls
			else {
				combo_pressed = 0;

				//handle new F200 style control scheme:
				if (_right_handed) {
			
					/* Delete windows */
					if (GetAsyncKeyState(GP2X_SELECT) < 0)
					{
						DeleteNonVitalWindows();
					}
					
					//DKS - new hotkey for landscaping tool
					if (GetAsyncKeyState(GP2X_START) < 0)
					{
						// simulate 'l' for landscaping hotkey
						uint32 tmpkey = 'L';
						HandleKeypress(tmpkey << 16);
						SDL_Delay(250);
					}	

					if( GetAsyncKeyState(!_controls_swapped ? GP2X_VOLUP : GP2X_L) < 0)
					{
						_left_button_down = true;
					}
					else
					{
						_left_button_down = false;
						_left_button_clicked = false;
					}
		
					if( GetAsyncKeyState(!_controls_swapped ? GP2X_VOLDOWN : GP2X_R) < 0) 
					{
						_right_button_down = true;
						_right_button_clicked = _right_button_down;
					}
					else
					{
						_right_button_down = false;
					}



					/* SHIFT pressed */
					if (GetAsyncKeyState(!_controls_swapped ? GP2X_R : GP2X_VOLUP) < 0)
					{
						_shift_pressed = true;
					}
					else
					{
						_shift_pressed = false;
					}	
			

					//DKS - new zoom-in/out combos:
					if ((GetAsyncKeyState(!_controls_swapped ? GP2X_L : GP2X_R) < 0) &&
						(GetAsyncKeyState(GP2X_UP) < 0))
					{
						// simulate Shift-F5 for zoom-in
						uint32 tmpkey = WKC_F5;
						tmpkey |= WKC_SHIFT;
						HandleKeypress(tmpkey << 16);
						SDL_Delay(250);
						combo_pressed = 1;
					} else if ((GetAsyncKeyState(!_controls_swapped ? GP2X_L : GP2X_R) < 0) &&
						(GetAsyncKeyState(GP2X_DOWN) < 0))
					{
						// simulate Shift-F6 for zoom-out
						uint32 tmpkey = WKC_F6;
						tmpkey |= WKC_SHIFT;
						HandleKeypress(tmpkey << 16);
						SDL_Delay(250);
						combo_pressed = 1;
					} else if ((GetAsyncKeyState(!_controls_swapped ? GP2X_L : GP2X_R) < 0) &&
						(GetAsyncKeyState(GP2X_LEFT) < 0))
					{
						// simulate 'a' for auto-rail hotkey
						uint32 tmpkey = 'A';
						HandleKeypress(tmpkey << 16);
						SDL_Delay(250);
						combo_pressed = 1;
					} else if ((GetAsyncKeyState(!_controls_swapped ? GP2X_L : GP2X_R) < 0) &&
						(GetAsyncKeyState(GP2X_RIGHT) < 0))
					{
						// simulate 'x' for transparent buildings hotkey
						uint32 tmpkey = 'X';
						HandleKeypress(tmpkey << 16);
						SDL_Delay(250);
						combo_pressed = 1;
					}

					//DKS - new hotkey, press all four scroll directions at once to bring up cheat menu
					if ((GetAsyncKeyState(GP2X_Y) < 0) && (GetAsyncKeyState(GP2X_X) < 0) 
						&& (GetAsyncKeyState(GP2X_A) < 0) && (GetAsyncKeyState(GP2X_B) < 0)) {
							// simulate 'Ctrl-Alt-c' for cheats hotkey
							uint32 tmpkey = 'C';
							tmpkey |= WKC_CTRL;
							tmpkey |= WKC_ALT;
							HandleKeypress(tmpkey << 16);
							SDL_Delay(250);
							combo_pressed = 1;
					}

					for(ikey = GP2X_A; ikey <= GP2X_Y; ikey++)
					{
						if( GetAsyncKeyState(ikey ) < 0 )
						{
							int _x, _y;
		
							_x = _cursor.pos.x;
							_y = _cursor.pos.y;
							switch(ikey ) // Cheap way of going through the dpad
							{
							case GP2X_Y:
								_y -= scrollrate;
								if( _y < 0 ) _y = 0;
								break;
							case GP2X_B:
								_x += scrollrate;
								if( _x > _cur_resolution[0] - 1 ) _x = _cur_resolution[0] - 1;
								break;
							case GP2X_X:
								_y += scrollrate;
								if( _y > _cur_resolution[1] - 1 ) _y = _cur_resolution[1] - 1;
							break;
							case GP2X_A:
								_x -= scrollrate;
								if( _x < 0 ) _x = 0;
								break;
							}
		
							if (_cursor.fix_at) 
							{
								int dx = _x - _cursor.pos.x;
								int dy = _y - _cursor.pos.y;
								if (dx != 0 || dy != 0) 
								{
									_cursor.delta.x += dx;
									_cursor.delta.y += dy;
								}
							}
							else 
							{
								_cursor.delta.x = _x - _cursor.pos.x;
								_cursor.delta.y = _y - _cursor.pos.y;
								_cursor.pos.x = _x;
								_cursor.pos.y = _y;
								_cursor.dirty = true;
							}
						}
					}
				
	
					if (!combo_pressed) {

						/* Handle CONTROL button (that also might move window under cursor depending on context) */
						if (GetAsyncKeyState(!_controls_swapped ? GP2X_L : GP2X_VOLDOWN) < 0)
						{
        					    Window *w;
					       	     w = FindWindowFromPt(_cursor.pos.x, _cursor.pos.y);
	
						        if (w->window_class != WC_MAIN_WINDOW &&
						                w->window_class != WC_SELECT_GAME &&
						                w->window_class != WC_MAIN_TOOLBAR &&
						                w->window_class != WC_STATUS_BAR &&
					        	        w->window_class != WC_TOOLBAR_MENU &&
					                	w->window_class != WC_TOOLTIPS) {
							                _left_button_down = true;
							                windowmove_pressed = true;
									scrollrate = _fast_cursorrate; // speed scrolling for window movement
					                } else {
		        			            _ctrl_pressed = true;
							}
						} else {
							if (windowmove_pressed) {
								windowmove_pressed = false;
								_left_button_down = false;
								_left_button_clicked = false;
								scrollrate = _slow_cursorrate;
							}
							_ctrl_pressed = false;
				                }	
				

//						/* CONTROL pressed */
//						if (GetAsyncKeyState(GP2X_L) < 0)
//						{
//							_ctrl_pressed = true;
//						}
//						else
//						{
//							_ctrl_pressed = false;
//						}
			
						_dirkeys =	(GetAsyncKeyState(GP2X_LEFT)	< 0	? 1 : 0) |
								(GetAsyncKeyState(GP2X_UP)	< 0	? 2 : 0) |
								(GetAsyncKeyState(GP2X_RIGHT)	< 0	? 4 : 0) |
								(GetAsyncKeyState(GP2X_DOWN)	< 0	? 8 : 0);
		 			} // !combo_pressed
					else {
						_dirkeys = 0;  // don't want new combos triggering other actions when pressed
					}

		
				} else {
					//left handed controls for those holding stylus in left hand			
						
					/* Delete windows */
					if (GetAsyncKeyState(GP2X_VOLUP) < 0)
					{
						DeleteNonVitalWindows();
					}
						
					// simulate 'l' for landscaping hotkey 
					if (GetAsyncKeyState(GP2X_VOLDOWN) < 0)
					{
						uint32 tmpkey = 'L';
						HandleKeypress(tmpkey << 16);
						SDL_Delay(250);
					}

					//DKS - these are reversed from L/R VOLUP/DOWN placement on purpose, It is better on my thumb 
					if( GetAsyncKeyState(!_controls_swapped ? GP2X_SELECT : GP2X_R) < 0)
					{
						_left_button_down = true;
					}
					else
					{
						_left_button_down = false;
						_left_button_clicked = false;
					}
		
					if( GetAsyncKeyState(!_controls_swapped ? GP2X_START : GP2X_L) < 0) 
					{
						_right_button_down = true;
						_right_button_clicked = _right_button_down;
					}
					else
					{
						_right_button_down = false;
					}
	
					//DKS - new zoom-in/out combos:
					if ((GetAsyncKeyState(!_controls_swapped ? GP2X_R : GP2X_L) < 0) &&
						(GetAsyncKeyState(GP2X_Y) < 0))
					{
						// simulate Shift-F5 for zoom-in
						uint32 tmpkey = WKC_F5;
						tmpkey |= WKC_SHIFT;
						HandleKeypress(tmpkey << 16);
						SDL_Delay(250);
						combo_pressed = 1;
					} else if ((GetAsyncKeyState(!_controls_swapped ? GP2X_R : GP2X_L) < 0) &&
						(GetAsyncKeyState(GP2X_X) < 0))
					{
						// simulate Shift-F6 for zoom-out
						uint32 tmpkey = WKC_F6;
						tmpkey |= WKC_SHIFT;
						HandleKeypress(tmpkey << 16);
						SDL_Delay(250);
						combo_pressed = 1;
					} else if ((GetAsyncKeyState(!_controls_swapped ? GP2X_R : GP2X_L) < 0) &&
						(GetAsyncKeyState(GP2X_A) < 0))
					{
						// simulate 'a' for auto-rail hotkey
						uint32 tmpkey = 'A';
						HandleKeypress(tmpkey << 16);
						SDL_Delay(250);
						combo_pressed = 1;
					} else if ((GetAsyncKeyState(!_controls_swapped ? GP2X_R : GP2X_L) < 0) &&
						(GetAsyncKeyState(GP2X_B) < 0))
					{
						// simulate 'x' for transparent buildings hotkey
						uint32 tmpkey = 'X';
						HandleKeypress(tmpkey << 16);
						SDL_Delay(250);
						combo_pressed = 1;
					}
					
					//DKS - new hotkey, press all four scroll directions at once to bring up cheat menu
					if ((GetAsyncKeyState(GP2X_Y) < 0) && (GetAsyncKeyState(GP2X_X) < 0) 
						&& (GetAsyncKeyState(GP2X_A) < 0) && (GetAsyncKeyState(GP2X_B) < 0)) {
							// simulate 'Ctrl-Alt-c' for cheats hotkey
							uint32 tmpkey = 'C';
							tmpkey |= WKC_CTRL;
							tmpkey |= WKC_ALT;
							HandleKeypress(tmpkey << 16);
							SDL_Delay(250);
							combo_pressed = 1;
					} 
 
					for(ikey = 0; ikey < 4; ikey++)
					{
						if( GetAsyncKeyState(ikey * 2) < 0 )
						{
							int _x, _y;
		
							_x = _cursor.pos.x;
							_y = _cursor.pos.y;
							switch(ikey * 2) // Cheap way of going through the dpad	
							{
								case GP2X_UP:
									_y -= scrollrate;
									if( _y < 0 ) _y = 0;
									break;
								case GP2X_RIGHT:
									_x += scrollrate;
									if( _x > _cur_resolution[0] - 1 ) _x = _cur_resolution[0] - 1;
									break;
								case GP2X_DOWN:
									_y += scrollrate;
									if( _y > _cur_resolution[1] - 1 ) _y = _cur_resolution[1] - 1;
									break;
								case GP2X_LEFT:
									_x -= scrollrate;
										if( _x < 0 ) _x = 0;
									break;
							}
		
							if (_cursor.fix_at) 
							{
								int dx = _x - _cursor.pos.x;
								int dy = _y - _cursor.pos.y;
									if (dx != 0 || dy != 0) 
								{
									_cursor.delta.x += dx;
										_cursor.delta.y += dy;
								}
							}
							else 
							{
								_cursor.delta.x = _x - _cursor.pos.x;
								_cursor.delta.y = _y - _cursor.pos.y;
								_cursor.pos.x = _x;
								_cursor.pos.y = _y;
								_cursor.dirty = true;
							}
						}
					}	
					
					/* Handle CONTROL button (that also might move window under cursor depending on context) */
					if (GetAsyncKeyState(!_controls_swapped ? GP2X_R : GP2X_SELECT) < 0)
					{
        				    Window *w;
				       	     w = FindWindowFromPt(_cursor.pos.x, _cursor.pos.y);
	
					        if (w->window_class != WC_MAIN_WINDOW &&
					                w->window_class != WC_SELECT_GAME &&
					                w->window_class != WC_MAIN_TOOLBAR &&
					                w->window_class != WC_STATUS_BAR &&
				        	        w->window_class != WC_TOOLBAR_MENU &&
				                	w->window_class != WC_TOOLTIPS) {
						                _left_button_down = true;
						                windowmove_pressed = true;
								scrollrate = _fast_cursorrate; // DKS - speed window movement  
				                } else {
	        			            _ctrl_pressed = true;
						}
					} else {
						if (windowmove_pressed) {
							windowmove_pressed = false;
							_left_button_down = false;
							_left_button_clicked = false;
							scrollrate = _slow_cursorrate;
						}
						_ctrl_pressed = false;
			                }	
	
//					/* CONTROL pressed */
//					if (GetAsyncKeyState(GP2X_R) < 0)
//					{
//						_ctrl_pressed = true;
//					}
//					else
//					{
//						_ctrl_pressed = false;
//					}
	
					if (!combo_pressed) {
						/* SHIFT pressed */
						if (GetAsyncKeyState(!_controls_swapped ? GP2X_L : GP2X_START) < 0)
						{
							_shift_pressed = true;
						}
						else
						{
							_shift_pressed = false;
						}
			
						_dirkeys =(GetAsyncKeyState(GP2X_A)	< 0	? 1 : 0) |
							(GetAsyncKeyState(GP2X_Y)	< 0	? 2 : 0) |
							(GetAsyncKeyState(GP2X_B)	< 0	? 4 : 0) |
							(GetAsyncKeyState(GP2X_X)	< 0	? 8 : 0);	
		 			} // !combo_pressed
					else {
						_dirkeys = 0;  // don't want new combos triggering other actions when pressed
					}

				}
			}	
		
#else
			_dirkeys =
				(keys[SDLK_LEFT]  ? 1 : 0) |
				(keys[SDLK_UP]    ? 2 : 0) |
				(keys[SDLK_RIGHT] ? 4 : 0) |
				(keys[SDLK_DOWN]  ? 8 : 0);
#endif
			GameLoop();

			_screen.dst_ptr = _sw_screen->pixels;
			UpdateWindows();
			if (++pal_tick > 4) {
				CheckPaletteAnim();
				pal_tick = 1;
			}

			DrawSurfaceToScreen();
		} else {
			//SDL_CALL SDL_Delay(1);
			_screen.dst_ptr = _sw_screen->pixels;
			DrawTextMessage();
			DrawMouseCursor();

			DrawSurfaceToScreen();
		}
	}
}

static bool SdlVideoChangeRes(int w, int h)
{
	return true; //CreateMainSurface(w, h);
}

static void SdlVideoFullScreen(bool full_screen)
{
#ifndef ARM
	_fullscreen = full_screen;
	GetVideoModes(); // get the list of available video modes
	if (_num_resolutions == 0 || !_video_driver->change_resolution(_cur_resolution[0], _cur_resolution[1])) {
		// switching resolution failed, put back full_screen to original status
		_fullscreen ^= true;
	}
#endif
}

const HalVideoDriver _sdl_video_driver = {
	SdlVideoStart,
	SdlVideoStop,
	SdlVideoMakeDirty,
	SdlVideoMainLoop,
	SdlVideoChangeRes,
	SdlVideoFullScreen,
};

#endif
